{
  "type": "record",
  "name": "DataPlatformInstanceKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a Dataset",
  "fields": [
    {
      "name": "platform",
      "type": "string",
      "doc": "Data platform urn associated with the instance",
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    },
    {
      "name": "instance",
      "type": "string",
      "doc": "Unique instance id"
    }
  ],
  "Aspect": {
    "name": "dataPlatformInstanceKey",
    "keyForEntity": "dataPlatformInstance",
    "entityCategory": "internal",
    "entityAspects": [
      "dataPlatformInstanceProperties",
      "ownership",
      "globalTags",
      "institutionalMemory",
      "deprecation",
      "status"
    ]
  }
}