{
  "type": "record",
  "name": "DataHubAccessTokenInfo",
  "namespace": "com.linkedin.pegasus2avro.access.token",
  "doc": "Information about a DataHub Access Token",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "doc": "User defined name for the access token if defined.",
      "Searchable": {
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "actorUrn",
      "type": "string",
      "doc": "Urn of the actor to which this access token belongs to.",
      "Searchable": {
        "fieldType": "URN"
      },
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    },
    {
      "name": "ownerUrn",
      "type": "string",
      "doc": "Urn of the actor which created this access token.",
      "Searchable": {
        "fieldType": "URN"
      },
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    },
    {
      "name": "createdAt",
      "type": "long",
      "doc": "When the token was created.",
      "Searchable": {
        "fieldType": "COUNT",
        "queryByDefault": false
      }
    },
    {
      "name": "expiresAt",
      "type": [
        "null",
        "long"
      ],
      "doc": "When the token expires.",
      "default": null,
      "Searchable": {
        "fieldType": "COUNT",
        "queryByDefault": false
      }
    },
    {
      "name": "description",
      "type": [
        "null",
        "string"
      ],
      "doc": "Description of the token if defined.",
      "default": null
    }
  ],
  "Aspect": {
    "name": "dataHubAccessTokenInfo"
  }
}